local mrotation_add = mrotation.add

local function to_radians( degrees )
	local degree_radian_constant = math.pi/180
	return degrees * degree_radian_constant
end

local function to_degrees( radians )
	local radian_degree_constant = 180/math.pi
	return radians * radian_degree_constant
end

-- This purely reads the animation values, please don't think this can actually parse model data.

SMDRig = SMDRig or class()
SMDRig.scale = 1
SMDRig.bone_mappings = {}

function SMDRig:add_bone( smd_bone, diesel_bone, position, rotation )
	SMDRig.bone_mappings[smd_bone] = {
		linkbone = diesel_bone,
		position = position,
		rotation = rotation
	}
end

SMDAnimation = SMDAnimation or class()
function SMDAnimation:init()
	self.frame_rate = 30
	self.frame_count = 0
	self.anim_length = 0
	self.smd = {
		bones = {},
		frames = {}
	}
	self.diesel = {
		bones = {},
		frames = {}
	}
end

function SMDAnimation:smd_to_diesel( rig )
	local last_bone_positions = {}
	local last_bone_rotations = {}

	for bone_index, bone_data in pairs(self.smd.bones) do
		self.diesel.bones[bone_index] = {
			name = rig and rig.bone_mappings[bone_data.name] and rig.bone_mappings[bone_data.name].linkbone or bone_data.name,
			parent = bone_data.parent
		}
	end

	for frame_number, frame_data in pairs(self.smd.frames) do
		self.diesel.frames[frame_number] = {}

		local function get_bone_position( bone_index )
			local bone_position = frame_data[bone_index].position or last_bone_positions[bone_index] or Vector3()
			last_bone_positions[bone_index] = bone_position

			return bone_position
		end

		local function get_bone_rotation( bone_index )
			local bone_rotation = frame_data[bone_index].rotation or last_bone_rotations[bone_index] or Rotation()
			last_bone_rotations[bone_index] = bone_rotation

			return bone_rotation
		end

		for bone_index, bone_frame_data in pairs(frame_data) do
			local base_pos = get_bone_position( bone_index )
			local base_rot = get_bone_rotation( bone_index )
			local bone_name = self.smd.bones[bone_index].name

			if rig and rig.bone_mappings[bone_name] then
				local offset_position = rig.bone_mappings[bone_name].position or Vector3()
				local offset_rotation = rig.bone_mappings[bone_name].rotation or Rotation()

				mrotation_add( base_rot, base_rot, offset_rotation )
				self.diesel.frames[frame_number][bone_index] = {
					position = base_pos + offset_position,
					rotation = base_rot
				}
			else
				self.diesel.frames[frame_number][bone_index] = {
					position = base_pos,
					rotation = base_rot
				}
			end
		end
	end
end

function SMDAnimation:output()
	local output = "---- ANIMATION OUTPUT ----\nversion 1\n"

	local function generate_nodes()
		local nodes = "nodes\n"
		for bone_index, bone_data in pairs(self.diesel.bones) do
			nodes = nodes .. tostring(bone_index) .. ' "' .. bone_data.name .. '" ' .. tostring(bone_data.parent) .. "\n"
		end
		nodes = nodes .. "end\n"

		output = output .. nodes
	end

	local function generate_frames()
		local frames = "skeleton\n"

		local frame_num = 0
		while ( frame_num <= self.frame_count - 1 ) do
			frames = frames .. "time " .. tostring(frame_num) .. "\n"

			local frame_data = self.diesel.frames[frame_num] or {}
			for bone_index, bone_data in pairs(frame_data) do
				local x = string.format("%.6f", bone_data.position.x)
				local y = string.format("%.6f", bone_data.position.y)
				local z = string.format("%.6f", bone_data.position.z)

				local rotation_z = string.format("%.6f", to_radians(bone_data.rotation:pitch()))
				local rotation_x = string.format("%.6f", to_radians(bone_data.rotation:yaw()))
				local rotation_y = string.format("%.6f", to_radians(bone_data.rotation:roll()))

				frames = frames .. tostring(bone_index) .. " " .. x .. " " .. y .. " " .. z .. " " .. rotation_x .. " " .. rotation_y .. " " .. rotation_z .. "\n"
			end

			frame_num = frame_num + 1
		end
		frames = frames .."end\n"

		output = output .. frames
	end

	generate_nodes()
	generate_frames()

	output = output .. "---- ANIMATION OUTPUT ----"

	local output_file = io.open("output.smd", "w")
	output_file:write(output)
	output_file:close()
end

SMDReader = SMDReader or {}

function SMDReader:parse_file( file_path )
	local smd_data = FileIO:ReadFrom( file_path )

	return self:parse_text( smd_data )
end

function SMDReader:parse_text( smd_input )
	local animation = SMDAnimation:new()

	local currently_nodes = false
	local currently_skeleton = false
	local current_frame = 0
	local highest_frame = 0

	for line in smd_input:gmatch("[^\r\n]+") do
		local line_data = {}
		for substring in line:gmatch("%S+") do
			table.insert(line_data, substring)
		end

		local command = line_data[1]

		if ( command == "nodes" ) then
			currently_nodes = true
		elseif ( command == "skeleton" ) then
			currently_skeleton = true
		elseif ( command == "end" ) then
			currently_nodes = false
			currently_skeleton = false
		elseif ( currently_nodes ) then
			local bone_index = tonumber(command)
			local bone_name = line_data[2]:gsub('"','')
			local bone_parent = tonumber(line_data[3]) or -1

			if bone_index then
				animation.smd.bones[bone_index] = { name = bone_name, parent = bone_parent }
			end
		elseif ( currently_skeleton and command == "time" ) then
			current_frame = tonumber(line_data[2]) or 0

			if current_frame > highest_frame then highest_frame = current_frame end

			animation.smd.frames[current_frame] = animation.smd.frames[current_frame] or {}
		elseif ( currently_skeleton ) then
			local bone_index = tonumber(command)

			local pos_x = tonumber(line_data[2]) or 0
			local pos_y = tonumber(line_data[3]) or 0
			local pos_z = tonumber(line_data[4]) or 0

			local rot_y = to_degrees( tonumber(line_data[5]) or 0 )
			local rot_r = to_degrees( tonumber(line_data[6]) or 0 )
			local rot_p = to_degrees( tonumber(line_data[7]) or 0 )

			animation.smd.frames[current_frame] = animation.smd.frames[current_frame] or {}
			animation.smd.frames[current_frame][bone_index] = {
				position = Vector3( pos_x, pos_y, pos_z ),
				rotation = Rotation( rot_p, rot_y, rot_r )
			}
		end
	end

	animation.frame_count = highest_frame
	animation.anim_length = highest_frame/animation.frame_rate

	return animation
end