Hooks:PostHook(UnitBase, "init", "SMD_Init", function( self, unit, update_enabled )
	self.update_enabled = update_enabled

	self._unit = unit

	self.smd_animations = {}
	self.smd_anim_current = nil
	self.smd_anim_start = 0
	self.smd_loop = false

	if not self.update_enabled then
		self._unit:set_extension_update_enabled(Idstring("base"), true)
		self.update_enabled = true
	end

	local old_update = self.update

	self.update = function( self, unit, t, dt )
		if old_update then
			old_update( self, unit, t, dt )
		end

		self:smd_update( unit, t, dt )
	end
end)

function UnitBase:add_smd_animation( id, data )
	self.smd_animations[id] = data
end

function UnitBase:play_smd_animation( anim_name, loop )
	if not self.smd_animations[anim_name] then return end

	self.smd_anim_current = anim_name
	self.smd_anim_start = managers.player:player_timer():time()
	self.smd_loop = loop or false

	if self._unit.anim_state_machine and self._unit:anim_state_machine() then
		local machine = self._unit:anim_state_machine()
		machine:set_enabled(false)
	end
end

function dump(o)
   if type(o) == 'table' then
      local s = '{ '
      for k,v in pairs(o) do
         if type(k) ~= 'number' then k = '"'..k..'"' end
         s = s .. '['..k..'] = ' .. dump(v) .. ','
      end
      return s .. '} '
   else
      return tostring(o)
   end
end

function UnitBase:record_smd_animation( time, frame_rate )
	self.smd_record_active = true
	self.smd_record_length = time
	self.smd_record_start = managers.player:player_timer():time()
	self.smd_record_end = self.smd_record_start + self.smd_record_length
	self.smd_record_rate = frame_rate
	self.smd_record_frame_count = self.smd_record_rate * self.smd_record_length
	
	self.smd_recorded_frames = {}

	if not self.smd_generated_rig then
		self.smd_generated_rig = {}

		local function index_by_name( name )
			for bone_index, bone in pairs(self.smd_generated_rig) do
				if ( bone.name == name ) then
					return bone_index
				end
			end
		end

		local bone_index = -1
		local function recurse_child( parent_object, object )
			bone_index = bone_index + 1

			self.smd_generated_rig[bone_index] = { ids = object:name(), name = object:name():key(), parent = -1 }
			if parent_object then
				self.smd_generated_rig[bone_index].parent = index_by_name(parent_object:name():key())
			end

			local child_objects = object:children()

			if child_objects then
				for _, child_object in ipairs(child_objects) do
					recurse_child( object, child_object )
				end
			end
		end

		recurse_child( nil, self._unit:orientation_object() )
	end
end

function UnitBase:stop_smd_animation()
	self.smd_anim_current = nil
	self.smd_loop = false

	if self._unit.anim_state_machine and self._unit:anim_state_machine() then
		local machine = self._unit:anim_state_machine()
		machine:set_enabled(true)
	end
end

local reverse_hash_table = {}
local mloor = math.floor
local meil = math.ceil
function UnitBase:smd_update( unit, t, dt )
	if self.smd_anim_current then
		local anim_data = self.smd_animations[self.smd_anim_current]

		local current_anim_time = t - self.smd_anim_start
		local total_anim_time = anim_data.anim_length

		local percentage = current_anim_time / total_anim_time
		local frame = (percentage * anim_data.frame_count) % anim_data.frame_count
		
		local lower_frame = mloor(frame)
		local upper_frame = meil(frame)
		local frame_percentage = frame - lower_frame

		local lower_frame_data = anim_data.diesel.frames[lower_frame] or {}
		local upper_frame_data = anim_data.diesel.frames[upper_frame] or {}

		local interp_frame_data = {}
		for bone_index, upper_object_frame_data in pairs(upper_frame_data) do
			if lower_frame_data[bone_index] then
				local lower_object_frame_data = lower_frame_data[bone_index]
				local object_frame_data = {
					position = Vector3(),
					rotation = Rotation()
				}
				if lower_object_frame_data.position then mvector3.lerp(object_frame_data.position, lower_object_frame_data.position, upper_object_frame_data.position, frame_percentage) end
				if lower_object_frame_data.rotation then mrotation.slerp(object_frame_data.rotation, lower_object_frame_data.rotation, upper_object_frame_data.rotation, frame_percentage) end

				interp_frame_data[bone_index] = object_frame_data 
			end
		end

		local bone_data = anim_data.diesel.bones
		for bone_index, object_frame_data in pairs(lower_frame_data) do
			local object_ids = Idstring(bone_data[bone_index].name)
			local object = unit:get_object(object_ids)

			if object then
				if interp_frame_data[bone_index] then
					local interp_frame_data = interp_frame_data[bone_index]

					if interp_frame_data.position then object:set_local_position( interp_frame_data.position ) end
					if interp_frame_data.rotation then object:set_local_rotation( interp_frame_data.rotation ) end
				else
					if object_frame_data.position then object:set_local_position( object_frame_data.position ) end
					if object_frame_data.rotation then object:set_local_rotation( object_frame_data.rotation ) end
				end
			end
		end

		if not self.smd_loop and current_anim_time > total_anim_time then
			self:stop_smd_animation()
		end
	end

	if self.smd_record_active then
		if ( t <= self.smd_record_end ) then
			local current_anim_time = t - self.smd_record_start

			local percentage = current_anim_time / self.smd_record_length
			local frame = mloor(percentage * self.smd_record_frame_count) % self.smd_record_frame_count

			if not self.smd_recorded_frames[frame] then
				log("Record Frame " .. tostring(frame))

				self.smd_recorded_frames[frame] = {}

				for bone_index, bone_data in pairs(self.smd_generated_rig) do
					local object_ids = bone_data.ids
					local object = unit:get_object(object_ids)

					if object then
						self.smd_recorded_frames[frame][bone_index] = {
							position = object:local_position(),
							rotation = object:local_rotation()
						}
					end
				end
			end
		else
			self.smd_record_active = false

			local output_anim = SMDAnimation:new()
			output_anim.frame_rate = self.smd_record_rate
			output_anim.anim_length = self.smd_record_length
			output_anim.frame_count = self.smd_record_rate * self.smd_record_length

			output_anim.diesel.bones = self.smd_generated_rig
			for bone_index, bone_data in pairs(output_anim.diesel.bones) do
				bone_data.name = reverse_hash_table[tostring(bone_data.name)] or tostring(bone_data.name)
			end

			output_anim.diesel.frames = self.smd_recorded_frames

			output_anim:output()
		end
	end
end

local things_to_reverse_hash = {
	"root_point",
	"Hips",
	"Spine",
	"Spine1",
	"Spine2",
	"Neck",
	"Head",
	"eyeAim",
	"LeftShoulder",
	"LeftArm",
	"LeftForeArm",
	"LeftForeArmRoll",
	"LeftHand",
	"LeftHandThumb1",
	"LeftHandThumb2",
	"LeftHandThumb3",
	"LeftHandIndex1",
	"LeftHandIndex2",
	"LeftHandIndex3",
	"LeftHandMiddle1",
	"LeftHandMiddle2",
	"LeftHandMiddle3",
	"LeftHandRing1",
	"LeftHandRing2",
	"LeftHandRing3",
	"LeftHandPinky1",
	"LeftHandPinky2",
	"LeftHandPinky3",
	"a_weapon_left_front",
	"RightShoulder",
	"RightArm",
	"RightForeArm",
	"RightForeArmRoll",
	"RightHand",
	"RightHandThumb1",
	"RightHandThumb2",
	"RightHandThumb3",
	"RightHandIndex1",
	"RightHandIndex2",
	"RightHandIndex3",
	"RightHandMiddle1",
	"RightHandMiddle2",
	"RightHandMiddle3",
	"RightHandRing1",
	"RightHandRing2",
	"RightHandRing3",
	"RightHandPinky1",
	"RightHandPinky2",
	"RightHandPinky3",
	"a_weapon_right_front",
	"aim",
	"e_taser",
	"BackPack1",
	"LeftUpLeg",
	"LeftLeg",
	"LeftFoot",
	"LeftToeBase",
	"RightUpLeg",
	"RightLeg",
	"RightFoot",
	"RightToeBase"
}

for index, thing in pairs(things_to_reverse_hash) do
	reverse_hash_table[Idstring(thing):key()] = thing
end